#include "stdlib.h"
#include "stdio.h"
#include "math.h"

#ifndef ISING_MODEL_DIM_X
#define ISING_MODEL_DIM_X 40 
#endif

#ifndef ISING_MODEL_DIM_Y
#define ISING_MODEL_DIM_Y 40 
#endif 

#ifndef ISING_MODEL_THRESH
#define ISING_MODEL_THRESH .50
#endif

#ifndef ISING_MODEL_ITER
#define ISING_MODEL_ITER 15
#endif

#ifndef ISING_MODEL_SEED
#define ISING_MODEL_SEED 1234567895
#endif

#ifndef ISING_MODEL_INFLUENCE
#define ISING_MODEL_INFLUENCE 100
#endif

#ifndef USE_ISING_MODEL_HOOK 
#define USE_ISING_MODEL_HOOK 1
#endif

#ifndef ISING_MODEL_2D_RULE
#define ISING_MODEL_2D_RULE 1
#endif

int int_4touple_max (
	int i1,
	int i2
);

int int_4touple_min (
	int i1,
	int i2
);

int int_4touple_modp (
	int i,
	int j
);

int int_4touple_wrap (
	int ival,
	int ilo,
	int ihi
);

void ising_model_2d_rule (
	int m,
	int n,
	int *c1,
	int *c5
);

int *ising_model_2d_init (
	int m,
	int n,
	double thresh,
	int *seed
);

void ising_model_2d_print_stdout(
	int m,
	int n,
	int *c1
);

void ising_model_2d_transition ( 
	int m,
	int n,
	int iterations,
	double *prob,
	double thresh,
	int *seed,
	int *c1
);

void double_8_uniform_distrib (
	int m,
	int n, 
	int *seed, 
	double *r
);

void gnugo_hook_influence_goban(
	int * c1,
	int **board, 
	int min, 
	int max
);


int * gnugo_hook_goban_to_im(
	int * board,
	int m,
	int n
);

#ifdef TEST
int main (
	int argc,
	char *argv[]
)
{
  int
	iterations = ISING_MODEL_ITER, 
	m = ISING_MODEL_DIM_X, 
	n = ISING_MODEL_DIM_Y, 
	seed = ISING_MODEL_SEED;
	
  double
	thresh = ISING_MODEL_THRESH;
	
  int *c1;
  
  double prob[5] =
  {
	  0.98,
	  0.85,
	  0.50,
	  0.15, 
	  0.02
  };
  
  c1 = ising_model_2d_init (
	  m,
	  n,
	  thresh,
	  &seed
  );
  
  ising_model_2d_transition (
	  m,
	  n,
	  iterations,
	  prob,
	  thresh,
	  &seed,
	  c1
  );
  
  ising_model_2d_print_stdout(
	  m,
	  n,
	  c1
  );
  
  free ( c1 );
  
  return 0;
}
#endif

void gnugo_hook_influence_goban(int * c1, int **board, int min, int max)
{
  int j;
  
  for ( j = min; j < max; j++ )	 
	  if ( c1[j] == -1 && rand() % ISING_MODEL_INFLUENCE == 1)
		  (*board)[j] = 0; /*GG_COLOR_BLACK*/
}

int * gnugo_hook_goban_to_im(int * board, int m, int n)
{
  int i, j;
  
  int *c1;
  double *r;

  r = malloc(m*n*sizeof(double));
  
  c1 = malloc(m*n*sizeof(int));

  for ( j = 0; j < n; j++ )
  {
    for ( i = 0; i < m; i++ )
    {
      if ( board[i+j*m] == 0 /*GG_COLOR_BLACK*/ )
	      c1[i+j*m] = -1;
      else
	      c1[i+j*m] = +1;
    }
  }
  
  free(r);

  return c1;
}

void ising_model_2d_print_stdout( int m, int n, int c1[] )
{
  int i, j;
  
  for ( i = 0; i < m ; i++ )
  {
      for ( j = 0; j < n; j++ )
      {
          printf("%d ", c1[i*m+j]);
      }
  }
}

int int_4touple_max ( int i1, int i2 )
{
  int value;
  
  if ( i2 < i1 ) 
	  value = i1;
  else 
	  value = i2;
  return value;
}

int int_4touple_min ( int i1, int i2 )
{
  int value;
  if ( i1 < i2 ) 
	  value = i1;
  else 
	  value = i2;
  return value;
}

int int_4touple_modp ( int i, int j )
{
  int value;
  if ( j == 0 )
	  exit ( 1 );
  value = i % j;
  if ( value < 0 )
	  value = value + abs ( j );
  return value;
}

int int_4touple_wrap ( int ival, int ilo, int ihi )
{
  int
	jhi,
	jlo, 
	value, 
	wide;

  jlo = int_4touple_min ( ilo, ihi );
  jhi = int_4touple_max ( ilo, ihi );

  wide = jhi + 1 - jlo;

  if ( wide == 1 ) 
	  value = jlo;
  else
	  value = jlo + int_4touple_modp ( ival - jlo, wide );

  return value;
}

void ising_model_2d_rule ( int m, int n, int c1[], int c5[] )
{
  int
	i, 
	j,
	im, 
	ip,
	jm, 
	jp;

  for ( j = 0; j < n; j++ )
  {
    jp = int_4touple_wrap ( j + ISING_MODEL_2D_RULE, 0, n - ISING_MODEL_2D_RULE );
    jm = int_4touple_wrap ( j - ISING_MODEL_2D_RULE, 0, n - ISING_MODEL_2D_RULE );
    for ( i = 0; i < m; i++ )
    {
      ip = int_4touple_wrap ( i + ISING_MODEL_2D_RULE, 0, m - ISING_MODEL_2D_RULE );
      im = int_4touple_wrap ( i - ISING_MODEL_2D_RULE, 0, m - ISING_MODEL_2D_RULE );
      c5[i+j*m] = c1[i+j*m] + c1[ip+j*m] + c1[im+j*m] + c1[i+jm*m] + c1[i+jp*m];
      if ( 0 < c1[i+j*m] ) 
	      c5[i+j*m] = ( 5 + c5[i+j*m] ) / 2;
      else 
	      c5[i+j*m] = ( 5 - c5[i+j*m] ) / 2;
    }
  }
  return;
}

/* rand case just for testing, not used in actual game **/
int *ising_model_2d_init ( int m, int n, double thresh, int *seed )
{
  int i, j;
  int *c1;
  double *r;

  r = malloc(m*n*sizeof(double));

  double_8_uniform_distrib ( m, n, seed, r );

  c1 = malloc(m*n*sizeof(int));

  for ( j = 0; j < n; j++ )
  {
    for ( i = 0; i < m; i++ )
    {
      if ( r[i+j*m] <= thresh )
	      c1[i+j*m] = -1;
      else
	      c1[i+j*m] = +1;
    }
  }
  free(r);

  return c1;
}

/* rand case just for testing, not used in actual game **/
void double_8_uniform_distrib ( int m, int n, int *seed, double r[] )
{
  int i4_huge = 2147483647;
  int k;
  int i, j;
  
  if ( *seed == 0 ) 
	  exit ( 1 );
  for ( j = 0; j < n; j++ )
  {
    for ( i = 0; i < m; i++ )
    {
      k = *seed / 127773;
      *seed = 16807 * ( *seed - k * 127773 ) - k * 2836;
      if ( *seed < 0 )
	      *seed = *seed + i4_huge;
      r[i+j*m] = ( double ) ( *seed ) * 4.656612875E-10;
    }
  }
  return;
}

void ising_model_2d_transition ( int m, int n, int iterations, double prob[],  double thresh, int *seed, int c1[] )
{
  int *c5;
  double *r;
  int step;
  int i, j;

  c5 = malloc(m*n*sizeof(int));
  
  r = malloc(m*n*sizeof(double));
  step = 0;

  for ( step = 1; step <= iterations; step++ )
  {
    ising_model_2d_rule ( m, n, c1, c5 );
    
    double_8_uniform_distrib ( m, n, seed, r );
    
    for ( j = 0; j < n; j++ )
	    for ( i = 0; i < m; i++ )
		    if ( r[i+j*m] < prob[c5[i+j*m]-1] )
			    c1[i+j*m] = - c1[i+j*m];
  }
  
  free(c5);
  free(r);

  return;
}